"""
The velociraptor module. 

More information is available in the documnetation.
"""

# First things first, we need to upgrade msun and mh from a symbol to a
# first-class unit. We also add a dimensionless unit for magnitudes.
import unyt

try:
    unyt.define_unit("msun", 1.0 * unyt.msun, tex_repr=r"M_\odot")
except RuntimeError:
    # We've already done that, oops.
    pass

try:
    unyt.define_unit("mh", 1.0 * unyt.hydrogen_mass, tex_repr=r"m_{\rm H}")
except RuntimeError:
    # We've already done that, oops.
    pass

try:
    unyt.define_unit("mag", 1.0 * unyt.dimensionless, tex_repr=r"mag")
except RuntimeError:
    # We've already done that, oops.
    pass


from velociraptor.catalogue.catalogue import VelociraptorCatalogue
from velociraptor.__version__ import __version__

from typing import Union, List


def load(
    filename: str,
    disregard_units: bool = False,
    registration_file_path: Union[List[str], str, None] = None,
    mask: slice = Ellipsis,
) -> VelociraptorCatalogue:
    """
    Loads a velociraptor catalogue, producing a VelociraptorCatalogue
    object.

    Parameters
    ----------

    filename: str
        The filename of your VELOCIraptor catalogue file (i.e.
        the path to the .properties file).

    disregard_units: bool, optional
        If ``True``, then disregard any additional units in the
        VELOCIraptor catalogues, and instead base everything on
        the 'base' units of velocity, length, and mass. In this
        case metallicities are left dimensionless. If you are
        using EAGLE data, you should set this to False, as the
        star formation rate units are presented in non-internal
        units.

    registration_file_path: Union[List[str], str], optional
        The filename of the derived quantities script(s) to register
        additional properties with the catalogue. This is an
        advanced feature. See the documentation for more details.

    mask: Union[None, NDArray[bool], int], optional
        If a boolean array is provided, it is used to mask all
        catalogue arrays. If an int is provided, catalogue arrays
        are masked to the single corresponding element.


    Returns
    -------

    VelociraptorCatalogue
        The VelociraptorCatalogue object that describes your
        .properties file.
    """

    catalogue = VelociraptorCatalogue(
        filename, disregard_units=disregard_units, mask=mask
    )

    if registration_file_path is not None:
        catalogue.register_derived_quantities(registration_file_path)

    return catalogue



"""
Tools for creating luminosity functions!
"""

import unyt
import numpy as np

from velociraptor.tools.labels import get_luminosity_function_label_no_units


def create_luminosity_function_given_bins(
    luminosities: unyt.unyt_array,
    bins: unyt.unyt_array,
    box_volume: unyt.unyt_quantity,
    minimum_in_bin: int = 3,
):
    """
    Creates a luminosity function (with equal width bins magnitudes) for you to plot.

    Parameters
    ----------

    luminosities: unyt.unyt_array
        The array that you want to create a luminosity function of (usually this is
        for example the stellar luminosities).

    bins: unyt.unyt_array
        The magnitude bin edges to use.

    unyt.unyt_quantity: box_volume
        The volume of the box such that we can return ``n / volume``.

    minimum_in_bin: int, optional
        The number of objects in a bin for it to be classed as valid. Bins
        with a number of objects smaller than this are not returned. By default
        this parameter takes a value of 3.


    Returns
    -------

    bin_centers: unyt.unyt_array
        The centers of the bins (taken to be the linear mean of the bin edges).

    luminosity_function: unyt.unyt_array
        The value of the luminosity function at the bin centers.

    error: unyt.unyt_array
        Scatter in the luminosity function (Poisson errors).

    """

    bins.convert_to_units(luminosities.units)

    # This is required to ensure that the luminosity function converges with bin width
    bin_width = bins[1] - bins[0]
    normalization_factor = 1.0 / (bin_width * box_volume)

    luminosity_function, _ = np.histogram(luminosities, bins)
    valid_bins = luminosity_function >= minimum_in_bin

    # Poisson sampling
    error = np.sqrt(luminosity_function)

    luminosity_function *= normalization_factor
    error *= normalization_factor

    bin_centers = 0.5 * (bins[1:] + bins[:-1])

    luminosity_function.name = get_luminosity_function_label_no_units("{}")
    bin_centers.name = luminosities.name

    return bin_centers[valid_bins], luminosity_function[valid_bins], error[valid_bins]


def create_luminosity_function(
    luminosities: unyt.unyt_array,
    lowest_magnitude: unyt.unyt_quantity,
    highest_magnitude: unyt.unyt_quantity,
    box_volume: unyt.unyt_quantity,
    n_bins: int = 25,
    minimum_in_bin: int = 3,
    return_bin_edges: bool = False,
):
    """
    Creates a luminosity function (with equal width bins in log M) for you to plot.

    Parameters
    ----------

    luminosities: unyt.unyt_array
        The array that you want to create a luminosity function of (usually this is
        for example stellar luminosities).

    lowest_magnitude: unyt.unyt_quantity
        the lowest magnitude edge of the bins
    
    highest_magnitude: unyt.unyt_quantity
        the highest magnitude edge of the bins

    box_volume: unyt.unyt_quantity
        The volume of the box such that we can return ``n / volume``.

    n_bins: unyt.unyt_array
        The number of equal log-width bins across the range to use.

    minimum_in_bin: int, optional
        The number of objects in a bin for it to be classed as valid. Bins
        with a number of objects smaller than this are not returned. By default
        this parameter takes a value of 3.

    return_bin_edges: bool, optional
        Return the bin edges used in the binning process? Default is False.

    Returns
    -------

    bin_centers: unyt.unyt_array
        The centers of the bins (taken to be the linear mean of the bin edges).

    luminosity_function: unyt.unyt_array
        The value of the luminosity function at the bin centers.

    error: unyt.unyt_array
        Scatter in the luminosity function (Poisson errors).

    bin_edges: unyt.unyt_array, optional
        Bin edges that were used in the binning process.

    """

    assert (
        luminosities.units == lowest_luminosity.units
        and lowest_luminosity.units == highest_luminosity.units
    ), "Please ensure that all luminosity quantities have the same units."

    bins = (
        np.linspace(lowest_magnitude, highest_magnitude, n_bins + 1)
        * luminosities.units
    )

    bin_centers, luminosity_function, error = create_luminosity_function_given_bins(
        luminosities=luminosities,
        bins=bins,
        box_volume=box_volume,
        minimum_in_bin=minimum_in_bin,
    )

    if return_bin_edges:
        return bin_centers, luminosity_function, error, bins
    else:
        return bin_centers, luminosity_function, error



"""
Velociraptor particles interaction functions. These allow users to 
extract the particles belonging to, e.g. a single halo.
"""

from velociraptor.particles.particles import VelociraptorGroups
from velociraptor.catalogue.catalogue import VelociraptorCatalogue

from typing import Union


def load_groups(
    filename, catalogue: Union[VelociraptorCatalogue, None] = None
) -> VelociraptorGroups:
    """
    Load the groups file, passed by its filename to this
    function. Passing the velociraptor catalogue to the
    catalogue argument allows for significantly more information
    to be extracted, and is highly recommended.
    """

    return VelociraptorGroups(filename, catalogue=catalogue)



"""
Main objects for the velociraptor reading library.

This is based upon the reading routines in the SWIFTsimIO library.
"""

import h5py
import unyt

import numpy as np

from typing import Union, Callable, List, Dict

from velociraptor.units import VelociraptorUnits
from velociraptor.catalogue.derived import DerivedQuantities
from velociraptor.catalogue.registration import global_registration_functions
from velociraptor.exceptions import RegistrationDoesNotMatchError


class VelociraptorFieldMetadata(object):
    """
    Metadata for a velociraptor field. Pass it a field path and a filename,
    and it will:

    + Use the registration functions to find the correct units and
      "fancy name".
    + Assign a proper snake_case_name to the dataset.
    """

    # Forward declarations for the field.

    # Is this a valid field?
    valid: bool = False
    # The fancy name of this field for plotting, provided by registration function
    name: str = ""
    # The snake case name of this field for use accessing the object.
    snake_case: str = ""
    # The unit of this field
    unit: unyt.Unit
    # The registartion function that matched with this field
    corresponding_registration_function_name: Union[str, None]
    corresponding_registration_function: Union[Callable, None]

    def __init__(
        self,
        filename,
        path: str,
        registration_functions: Dict[str, Callable],
        units: VelociraptorUnits,
    ):
        """
        I take in:

        + filename of the velociraptor properties file
        + path of the field you wish me to look at
        + registration_functions a list of callables with the registration
          function signature (see registration.py or the documentation)
        + units, a pointer or copy of the unit system associated with this file.
        """

        # Filename not currently used but may be required later on if
        # actual field metadata is included in the velociraptor properties files
        self.filename = filename
        self.path = path
        self.registration_functions = registration_functions
        self.units = units

        self.register_field_properties()

        return

    def register_field_properties(self):
        """
        Registers the field properties using the registration functions.
        """

        for reg_name, reg in self.registration_functions.items():
            try:
                self.unit, self.name, self.snake_case = reg(
                    field_path=self.path, unit_system=self.units
                )
                self.valid = True
                self.corresponding_registration_function = reg
                self.corresponding_registration_function_name = reg_name
            except RegistrationDoesNotMatchError:
                continue

        return


def generate_getter(filename, name: str, field: str, full_name: str, unit):
    """
    Generates a function that:

    a) If self._`name` exists, return it
    b) If not, open `filename`
    c) Reads filename[`field`]
    d) Set self._`name`
    e) Return self._`name`.

    Takes:

    + filename, the filename of the hdf5 file
    + name, the snake_case name of the property
    + field, the field in the hdf5 file corresponding to this property
    + full_name, the fancy printing name for this quantity (registered to array.name)
    + unit, the unyt unit corresponding to this value
    """

    def getter(self):
        current_value = getattr(self, f"_{name}")

        if current_value is not None:
            return current_value
        else:
            with h5py.File(filename, "r") as handle:
                try:
                    mask = getattr(self, "mask")
                    setattr(
                        self, f"_{name}", unyt.unyt_array(handle[field][mask], unit)
                    )
                    getattr(self, f"_{name}").name = full_name
                    getattr(self, f"_{name}").file = filename
                except KeyError:
                    print(f"Could not read {field}")
                    return None

        return getattr(self, f"_{name}")

    return getter


def generate_setter(name: str):
    """
    Generates a function that sets self._name to the value that is passed to it.
    """

    def setter(self, value):
        setattr(self, f"_{name}", value)

        return

    return setter


def generate_deleter(name: str):
    """
    Generates a function that destroys self._name (sets it back to None).
    """

    def deleter(self):
        current_value = getattr(self, f"_{name}")
        del current_value
        setattr(self, f"_{name}", None)

        return

    return deleter


def generate_sub_catalogue(
    filename,
    registration_name: str,
    registration_function: Callable,
    units: VelociraptorUnits,
    field_metadata: List[VelociraptorFieldMetadata],
    mask: slice = Ellipsis,
):
    """
    Generates a sub-catalogue object with the correct properties set.
    This is required as we can add properties to a class, but _not_
    to an object dynamically.

    So, here, we initialise the metadata, create a _copy_ of the
    __VelociraptorSubCatlaogue class, and then add all of our properties
    to that _class_ before instantiating it with the metadata.
    """

    # This creates a _copy_ of the _class_, not object.
    this_sub_catalogue_bases = (__VelociraptorSubCatalogue, object)
    this_sub_catalogue_dict = {}

    valid_sub_paths = []

    for metadata in field_metadata:
        valid_sub_paths.append(metadata.snake_case)

        this_sub_catalogue_dict[metadata.snake_case] = property(
            generate_getter(
                filename,
                metadata.snake_case,
                metadata.path,
                metadata.name,
                metadata.unit,
            ),
            generate_setter(metadata.snake_case),
            generate_deleter(metadata.snake_case),
        )

        this_sub_catalogue_dict[f"_{metadata.snake_case}"] = None

    ThisSubCatalogue = type(
        f"Dynamic_{registration_name}_VelociraptorCatalogue",
        this_sub_catalogue_bases,
        this_sub_catalogue_dict,
    )

    # Finally, we can actually create an instance of our new class.
    catalogue = ThisSubCatalogue(filename=filename, mask=mask)
    catalogue.valid_sub_paths = valid_sub_paths

    return catalogue


class __VelociraptorSubCatalogue(object):
    """
    A velociraptor mini-catalogue, containing the only the information from one
    registration function. This allows us to separate the top-level variables
    into more manageable chunks.

    Do not directly instantiate this class, you should use generate_sub_catalogue.
    This is called in VelociraptorCatalogue.
    """

    # The valid paths contained within
    valid_sub_paths: List[str]

    def __init__(self, filename, mask=Ellipsis):
        self.filename = filename
        self.mask = mask

        return

    def __str__(self):
        return f"Contains the following fields: {', '.join(self.valid_sub_paths)}"

    def __repr__(self):
        return str(self)


class VelociraptorCatalogue(object):
    """
    A velociraptor dataset, containing information that has correct units
    and are easily accessible through snake_case names.
    """

    # Top-level definitions for autocomplete
    registration_functions: Union[List[Callable], None]

    def __init__(
        self,
        filename: str,
        disregard_units: bool = False,
        extra_registration_functions: Union[None, Dict[str, Callable]] = None,
        mask: slice = Ellipsis,
    ):
        """
        Initialise the velociraptor catalogue with all of the available
        datasets. This class should never be instantiated manually and should
        always be handled through the generate_catalogue function.

        Parameters
        ----------

        filename: str
            File path to the VELOCIraptor .properties file that you wish to open.

        disregard_units: bool, optional
            If ``True``, then disregard any additional units in the
            VELOCIraptor catalogues, and instead base everything on
            the 'base' units of velocity, length, and mass. In this
            case metallicities are left dimensionless. If you are
            using EAGLE data, you should set this to False, as the
            star formation rate units are presented in non-internal
            units.

        extra_registration_functions:  Union[None, Dict[str, Callable]], optional
            Any additional registration functions that you wish to use. This
            should be a dictionary of strings pointing to callables, which
            conform to the registration function API. This is an advanced
            feature.

        mask: slice, optional
            If a boolean array is provided, it is used to mask all catalogue
            arrays. If an int is provided, catalogue arrays are masked to the
            single corresponding element. Default: Ellipsis (``...``).
        """
        self.filename = filename
        self.disregard_units = disregard_units
        self.extra_registration_functions = extra_registration_functions
        self.mask = mask

        self.get_units()
        self.extract_properties_from_units()

        self.__register_extra_registration_functions()

        self.__create_sub_catalogues()

        return

    def __str__(self):
        """
        Prints out some more useful information, rather than just
        the memory location.
        """

        if self.mask is Ellipsis:
            return (
                f"Velociraptor catalogue at {self.filename}. "
                "Contains the following field collections: "
                f"{', '.join(self.valid_field_metadata.keys())}"
            )
        else:
            return (
                f"Masked velociraptor catalogue at {self.filename}. "
                "Contains the following field collections: "
                f"{', '.join(self.valid_field_metadata.keys())}"
            )

    def __repr__(self):
        return str(self)

    def get_units(self):
        """
        Gets the units instance from the file properties.
        """

        self.units = VelociraptorUnits(
            self.filename, disregard_units=self.disregard_units
        )

        return self.units

    def __register_extra_registration_functions(self):
        """
        Sets the self.registration_functions attribute such that it includes
        both the globals and any user-provided extra values.
        """
        if self.extra_registration_functions is not None:
            self.registration_functions = {
                **self.extra_registration_functions,
                **global_registration_functions,
            }
        else:
            self.registration_functions = global_registration_functions

        return

    def extract_properties_from_units(self):
        """
        Use the self.units object to extract interesting parameters
        that should be visible from the top-level.
        """

        # Register some properties from the units that may be useful outside
        properties = ["a", "scale_factor", "z", "redshift"]

        for property in properties:
            setattr(self, property, getattr(self.units, property))

        return

    def __create_sub_catalogues(self):
        """
        Creates the sub-catalogues by instantiating many different versions
        of the __VelociraptorSubCatalogue. Each sub-catalogue corresponds to
        the output of a single registration function.
        """

        # First load all field names from the HDF5 file so that they can be parsed.

        with h5py.File(self.filename, "r") as handle:
            field_paths = list(handle.keys())

        # Now build metadata:
        self.valid_field_metadata = {
            reg: [] for reg in self.registration_functions.keys()
        }
        self.invalid_field_paths = []

        for path in field_paths:
            metadata = VelociraptorFieldMetadata(
                self.filename, path, self.registration_functions, self.units
            )

            if metadata.valid:
                self.valid_field_metadata[
                    metadata.corresponding_registration_function_name
                ].append(metadata)
            else:
                self.invalid_field_paths.append(path)

        # For each registration function, we create a dynamic sub-class that
        # contains only that information - otherwise the namespace of the
        # VelociraptorCatalogue is way too crowded.
        for attribute_name, field_metadata in self.valid_field_metadata.items():
            setattr(
                self,
                attribute_name,
                generate_sub_catalogue(
                    filename=self.filename,
                    registration_name=attribute_name,  # This ensures each class has a unique name
                    registration_function=self.registration_functions[attribute_name],
                    units=self.units,
                    field_metadata=field_metadata,
                    mask=self.mask,
                ),
            )

        return

    def register_derived_quantities(
        self, registration_file_path: Union[List[str], str]
    ) -> None:
        """
        Register any required derived quantities. These will
        be available through `catalogue.derived_quantities.{your_names}`.

        For more information on this feature, see the documentation of
        :class:`velociraptor.catalogue.derived.DerivedQuantities`.

        Parameters
        ----------

        registration_file_path: Union[List[str], str]
            Path to the python source file(s) that contain(s) the code to
            register the additional derived quantities.
        """

        self.derived_quantities = DerivedQuantities(registration_file_path, self)

        return

    @property
    def centrals(self):
        if hasattr(self.structure_type, "structuretype"):
            return self.structure_type.structuretype == 10
        else:
            return np.s_[:]

    @property
    def satellites(self):
        if hasattr(self.structure_type, "structuretype"):
            return self.structure_type.structuretype != 10
        else:
            return np.s_[:]



"""
Autoplotter module. Contains the functionality to use the library
to automatically make plots of variable x against variable y, by
reading from a yaml file.

This yaml file has the following format:

.. code-block:: yaml

   output_filename:
     # Global plot quantities
     type: scatter / histogram / cumulative_histogram / 2dhistogram / massfunction / adaptivemassfunction
     select_structure_type: number of substructure to select (e.g. 10 for centrals)
     exclude_structure_type: number of substructure to exclude (e.g. 10 for non-centrals). Cannot be a number equal to select_structure_type.
     selection_mask: "quantity.name" - a boolean quantity to select the items you wish to plot.
     number_of_bins: number of bins in the (background) histogram or massfunc
     min_num_points_highlight: Minimum number of data points (with the largest values of x) to highlight in mean- and median-line plots (default: 10)
     reverse_cumsum: reverse the cumulative sum in cumulative histogram plots? Choose true or false
     comment: "An extra string to be placed on the plot"
     legend_loc: Location of legend (string). One of:
            "upper right",
            "upper left",
            "lower left",
            "lower right",
            "right",
            "lower center",
            "upper center",
            "center",
     redshift_loc: Location of z=x a=y string. Same vailidity as legend_loc.
     comment_loc: Location of extra comment string. Same validity as legend_loc
     # Horizontal quantity
     x:
       quantity: "quantity.name" (e.g. masses.mass_200crit)
       units: units to plot in (e.g. Solar_Mass)
       start: horizontal axis start (in units above)
       end: horizontal axis to end (in units above)
       log: true/false (true by default), do you want a log axis?
       label_override: override the label with your own choice
       # if hist or massfunc, in units above
       start_bin: number to start binning at (default is start)
       end_bin: number to end binning at (default is end)
       # Shade below or above a value, to show you do not trust this region
       shade:
         below: number (in same units)
         above: number
     # Vertical quantity - only need units for massfunction
     y:
       quantity: "quantity.name" (e.g. masses.mass_200crit)
       units: units to plot in (e.g. Solar_Mass)
       start: vertical axis start (in units above)
       end: vertical axis to end (in units above)
       log: true/false (true by default), do you want a log axis?
       label_override: override the label with your own choice
       # Shade below or above a value, to show you do not trust this region
       shade:
         below: number (in same units)
         above: number
     # Lines that you can plot on top - line_type can be median or mean
     line_type: 
       plot: true/false (true by default) actually plot this?
       log: true/false (true by default) use log bins?
       scatter: "none", "errorbar", "errorbar_both", or "shaded". Defaults to shaded
       number_of_bins: number of bins for median line, different from number_of_bins above
       adaptive: use adaptive binning?
       start: value to start binning at
         value: start value
         units: start value units
       end:
         value: end value
         units: end value units
       # Restriction of binning vertically
       lower:
         value: lower value for median/mean line
         units: lower value units
       upper:
         value: upper value for median/mean line
         units: upper value units
     # Generic metadata for plot
     metadata:
       title: title for the plot, should you want one
       caption: caption for the plot
       section: section the plot 'lives in' for pipeline usage.
       observational_data_bracket_width: 0.1 - the width in log(a) of the window to plot observational
       data (defined below) within.
     # Comparison data to plot on top
     observational_data:
       - filename: filename of data to plot on this plot (use ObservationalData written to disk)
       - filename: second filename for observational data
       - filename: ... you can plot as many of these as you'd like
"""



"""
Objects for handling and plotting mean and median lines.
"""

from unyt import unyt_quantity, unyt_array
from numpy import logspace, linspace, log10, logical_and, isnan, sqrt, logical_or
from typing import Dict, Union, Tuple, List
from matplotlib.pyplot import Axes
from matplotlib.transforms import blended_transform_factory

import velociraptor.tools.lines as lines
from velociraptor.tools.mass_functions import (
    create_mass_function_given_bins,
    create_adaptive_mass_function,
)
from velociraptor.tools.luminosity_functions import (
    create_luminosity_function_given_bins,
)
from velociraptor.tools.histogram import create_histogram_given_bins
from velociraptor.tools.adaptive import create_adaptive_bins
from velociraptor.autoplotter.box_size_correction import VelociraptorBoxSizeCorrection

valid_line_types = [
    "median",
    "mean",
    "mass_function",
    "luminosity_function",
    "histogram",
    "cumulative_histogram",
    "adaptive_mass_function",
]


class VelociraptorLine(object):
    """
    A median or mean line and all the information that is
    required for this (e.g. bins, log space, etc.)
    """

    # Forward declarations
    # Actually plot this line?
    plot: bool
    # Is a median, mass function, or a mean line?
    median: bool
    mean: bool
    mass_function: bool
    luminosity_function: bool
    histogram: bool
    cumulative_histogram: bool
    adaptive_mass_function: bool
    adaptive_luminosity_function: bool
    # Create bins in logspace?
    log: bool
    # Binning properties
    number_of_bins: int
    start: unyt_quantity
    end: unyt_quantity
    lower: unyt_quantity
    upper: unyt_quantity
    # Use adaptive binning?
    adaptive: bool
    bins: unyt_array = None
    # Scatter can be: "none", "errorbar", or "shaded"
    scatter: str
    # Box size correction?
    box_size_correction: Union[None, VelociraptorBoxSizeCorrection]
    # Output: centers, values, scatter, additional_x, additional_y - initialised here
    # to prevent crashes in other code.
    output: Tuple[unyt_array] = (
        unyt_array([]),
        unyt_array([]),
        unyt_array([]),
        unyt_array([]),
        unyt_array([]),
    )

    def __init__(
        self,
        line_type: str,
        line_data: Dict[str, Union[Dict, str]],
        box_size_correction: Union[None, VelociraptorBoxSizeCorrection] = None,
    ):
        """
        Initialise a line with data from the yaml file.
        """

        self.line_type = line_type
        self._parse_line_type()

        self.data = line_data
        self._parse_data()

        self.box_size_correction = box_size_correction

        return

    def _parse_line_type(self):
        """
        Parse the line type to a boolean.
        """

        # TODO: Use centralised metadata for this list.
        for line_type in valid_line_types:
            setattr(self, line_type, self.line_type == line_type)

        return

    def _parse_data(self):
        """
        Parse the line data from the dictionary and set defaults.
        """

        self.plot = bool(self.data.get("plot", True))
        self.log = bool(self.data.get("log", True))
        self.number_of_bins = int(self.data.get("number_of_bins", 25))
        self.scatter = str(self.data.get("scatter", "shaded"))
        self.adaptive = bool(self.data.get("adaptive", False))

        if self.scatter not in ["none", "errorbar", "shaded"]:
            self.scatter = "shaded"

        try:
            self.start = unyt_quantity(
                float(self.data["start"]["value"]), units=self.data["start"]["units"]
            )
        except KeyError:
            self.start = unyt_quantity(0.0)

        try:
            self.end = unyt_quantity(
                float(self.data["end"]["value"]), units=self.data["end"]["units"]
            )
        except KeyError:
            self.end = unyt_quantity(0.0)

        try:
            self.lower = unyt_quantity(
                float(self.data["lower"]["value"]), units=self.data["lower"]["units"]
            )
        except KeyError:
            self.lower = None

        try:
            self.upper = unyt_quantity(
                float(self.data["upper"]["value"]), units=self.data["upper"]["units"]
            )
        except KeyError:
            self.upper = None

        return

    def generate_bins(self, values=None):
        """
        Generates the required bins. If we request adaptive bins, this is a little
        more complicated and requires the values along the horizontal axis.
        """

        if values is not None and self.adaptive:
            self.end.convert_to_units(values.units)
            self.start.convert_to_units(self.end.units)

            bin_centers, bin_edges = create_adaptive_bins(
                values=values,
                lowest_value=self.start,
                highest_value=self.end,
                base_n_bins=self.number_of_bins,
                logarithmic=self.log,
                stretch_final_bin="mass_function" in self.line_type,
            )

            self.bins = bin_edges
        else:
            # Assert these are in the same units just in case
            self.start.convert_to_units(self.end.units)

            if self.log:
                # Need to strip units, unfortunately
                self.bins = unyt_array(
                    logspace(
                        log10(self.start.value),
                        log10(self.end.value),
                        self.number_of_bins,
                    ),
                    units=self.start.units,
                )
            else:
                # Can get away with this one without stripping
                self.bins = linspace(self.start, self.end, self.number_of_bins)

        return

    def create_line(
        self,
        x: unyt_array,
        y: unyt_array,
        box_volume: Union[None, unyt_quantity] = None,
        reverse_cumsum: bool = False,
        minimum_additional_points: int = 0,
    ):
        """
        Creates the line!

        Parameters
        ----------

        x: unyt_array
            Horizontal axis data

        y: unyt_array
            Vertical axis data

        box_volume: Union[None, unyt_quantity]
            Box volume for the simulation, required for mass functions. Should
            have associated volume units. Generally this is given as a comoving
            quantity.

        reverse_cumsum: bool
            A boolean deciding whether to reverse the cumulative sum. If false,
            the sum is computed from low to high values (along the X-axis). Relevant
            only for cumulative histogram lines. Default is false.

        minimum_additional_points: int, optional
            Minimum number of additional data points with the highest values of x to
            show in the median-line or mean-line plots.

        Returns
        -------

        output: Tuple[unyt_array]
            A five-length (mean, median lines) or three-length (mass_function,
            histogram, cumulative_histogram) tuple of unyt arrays that takes the
            following form: (bin centers, vertical values, vertical scatter,
            additional_x [optional], additional_y [optional]).
        """

        if self.bins is None:
            self.generate_bins(values=x)
        else:
            self.bins.convert_to_units(x.units)

        self.output = None

        masked_x = x
        masked_y = y

        # filter out NaN values to prevent crashes if all x values are NaN
        if masked_y is not None:
            mask = isnan(x) | isnan(y)
            masked_x = masked_x[~mask]
            masked_y = masked_y[~mask]
        else:
            mask = isnan(x)
            masked_x = masked_x[~mask]

        if self.lower is not None:
            self.lower.convert_to_units(y.units)
            mask = masked_y > self.lower
            masked_x = masked_x[mask]
            masked_y = masked_y[mask]

        if self.upper is not None:
            self.upper.convert_to_units(y.units)
            mask = masked_y < self.upper
            masked_x = masked_x[mask]
            masked_y = masked_y[mask]

        if self.median:
            self.output = lines.binned_median_line(
                x=masked_x,
                y=masked_y,
                x_bins=self.bins,
                return_additional=True,
                minimum_additional_points=minimum_additional_points,
            )
        elif self.mean:
            self.output = lines.binned_mean_line(
                x=masked_x,
                y=masked_y,
                x_bins=self.bins,
                return_additional=True,
                minimum_additional_points=minimum_additional_points,
            )
        elif self.mass_function:
            mass_function_output = create_mass_function_given_bins(
                masked_x, self.bins, box_volume=box_volume
            )
            if self.box_size_correction is not None:
                mass_function_output = self.box_size_correction.apply_mass_function_correction(
                    mass_function_output
                )
            self.output = (
                *mass_function_output,
                unyt_array([], units=mass_function_output[0].units),
                unyt_array([], units=mass_function_output[1].units),
            )
        elif self.luminosity_function:
            luminosity_function_output = create_luminosity_function_given_bins(
                masked_x, self.bins, box_volume=box_volume
            )
            self.output = (
                *luminosity_function_output,
                unyt_array([], units=luminosity_function_output[0].units),
                unyt_array([], units=luminosity_function_output[1].units),
            )
        elif self.histogram:
            histogram_output = create_histogram_given_bins(
                masked_x, self.bins, box_volume=box_volume
            )
            self.output = (
                *histogram_output,
                unyt_array([], units=histogram_output[0].units),
                unyt_array([], units=histogram_output[1].units),
            )
        elif self.cumulative_histogram:
            histogram_output = create_histogram_given_bins(
                masked_x,
                self.bins,
                box_volume=box_volume,
                cumulative=True,
                reverse=reverse_cumsum,
            )
            self.output = (
                *histogram_output,
                unyt_array([], units=histogram_output[0].units),
                unyt_array([], units=histogram_output[1].units),
            )
        elif self.adaptive_mass_function:
            *mass_function_output, self.bins = create_adaptive_mass_function(
                masked_x,
                lowest_mass=self.start,
                highest_mass=self.end,
                box_volume=box_volume,
                return_bin_edges=True,
            )
            if self.box_size_correction is not None:
                mass_function_output = self.box_size_correction.apply_mass_function_correction(
                    mass_function_output
                )
            self.output = (
                *mass_function_output,
                unyt_array([], units=mass_function_output[0].units),
                unyt_array([], units=mass_function_output[1].units),
            )
        else:
            self.output = None

        return self.output

    def highlight_data_outside_domain(
        self,
        ax: Axes,
        x: unyt_array,
        y: unyt_array,
        color: str,
        x_lim: List,
        y_lim: List,
    ) -> None:

        """
        Add arrows to the plot for each data point residing outside the plot's domain.
        The arrows indicate where the missing points are. For a given missing data point
        with its Y(X) coordinate outside the Y(X)-axis range, the corresponding arrow
        will have the same X(Y) coordinate and point to the direction where the missing
        point is. If a data point happens to lie outside both the X-axis range and
        Y-axis range, then a diagonal arrow is drawn.

        Parameters
        ----------

        ax: Axes
            An object of axes where to draw the arrows

        x: unyt_array
            Horizontal axis data

        y: unyt_array
            Vertical axis data

        color: str
            Color of the arrows that this function will draw. The color should be the
            same as the color of the (missing) data points.

        x_lim: List
            A 2-length list containing the lower and upper limits of the X-axis range.

        y_lim: List
            A 2-length list containing the lower and upper limits of the Y-axis range.
        """

        # Additional check to ensure all provided data points are good
        if not isnan(x).any() and not isnan(y).any():

            # Arrow parameters
            arrow_length = 0.07
            distance_from_edge = 0.01
            arrow_style = "->"

            # Split data into three categories (along X axis)
            below_x_range = x < x_lim[0]
            above_x_range = x > x_lim[1]
            within_x_range = logical_and(x >= x_lim[0], x <= x_lim[1])

            # Split data into three categories (along Y axis)
            below_y_range = y < y_lim[0]
            above_y_range = y > y_lim[1]
            within_y_range = logical_and(y >= y_lim[0], y <= y_lim[1])

            # First, find all data points that are outside the Y-axis range and within
            # X-axis range
            below_y_within_x = logical_and(below_y_range, within_x_range)
            above_y_within_x = logical_and(above_y_range, within_x_range)

            # X coordinates of the data points whose Y coordinates are outside the
            # Y-axis range
            x_down_list = x[below_y_within_x]
            x_up_list = x[above_y_within_x]

            # Use figure's data coordinates along the X axis and relative coordinates
            # along the Y axis.
            tform_x = blended_transform_factory(ax.transData, ax.transAxes)

            # Draw arrows pointing downwards
            for x_down in x_down_list:
                # We are using 'ax.annotate' instead of 'ax.arrow' because we want the
                # arrow's head and tail to have the same size regardless of what the
                # axes aspect ratio is or whether the plot is in logarithmic or linear
                # scale.
                ax.annotate(
                    "",
                    xytext=(x_down, arrow_length + distance_from_edge),
                    textcoords=tform_x,
                    xy=(x_down, distance_from_edge),
                    xycoords=tform_x,
                    arrowprops=dict(color=color, arrowstyle=arrow_style),
                )

            # Draw arrows pointing upwards
            for x_up in x_up_list:
                ax.annotate(
                    "",
                    xytext=(x_up, 1.0 - arrow_length - distance_from_edge),
                    textcoords=tform_x,
                    xy=(x_up, 1.0 - distance_from_edge),
                    xycoords=tform_x,
                    arrowprops=dict(color=color, arrowstyle=arrow_style),
                )

            # Next, find all data points that are outside the X-axis range and
            # within Y-axis range
            below_x_within_y = logical_and(below_x_range, within_y_range)
            above_x_within_y = logical_and(above_x_range, within_y_range)

            # Y coordinates of the data points whose X coordinates are outside the
            # X-axis range
            y_left_list = y[below_x_within_y]
            y_right_list = y[above_x_within_y]

            # Use figure's data coordinates along the Y axis and relative coordinates
            # along the X axis.
            tform_y = blended_transform_factory(ax.transAxes, ax.transData)

            # Draw arrows pointing leftwards
            for y_left in y_left_list:
                ax.annotate(
                    "",
                    xytext=(arrow_length + distance_from_edge, y_left),
                    textcoords=tform_y,
                    xy=(distance_from_edge, y_left),
                    xycoords=tform_y,
                    arrowprops=dict(color=color, arrowstyle=arrow_style),
                )

            # Draw arrows pointing rightwards
            for y_right in y_right_list:
                ax.annotate(
                    "",
                    xytext=(1.0 - arrow_length - distance_from_edge, y_right),
                    textcoords=tform_y,
                    xy=(1.0 - distance_from_edge, y_right),
                    xycoords=tform_y,
                    arrowprops=dict(color=color, arrowstyle=arrow_style),
                )

            # Finally, handle the points that are both outside the X and Y axis range
            outside_plot = logical_and(
                logical_or(below_y_range, above_y_range),
                logical_or(below_x_range, above_x_range),
            )
            x_outside_list, y_outside_list = x[outside_plot], y[outside_plot]

            for x_outside, y_outside in zip(x_outside_list, y_outside_list):

                # Unlike vertical and horizontal arrows, diagonal arrows extend both
                # in X and Y directions. We account for it by dividing the length of
                # diagonal arrow along each dimension by \sqrt(2).
                arrow_proj_length = arrow_length / sqrt(2.0)

                # Find the correct position of the arrow on the plot
                if x_lim[0] > x_outside:
                    arrow_start_x = arrow_proj_length + distance_from_edge
                    arrow_end_x = distance_from_edge
                else:
                    arrow_start_x = 1.0 - arrow_proj_length - distance_from_edge
                    arrow_end_x = 1.0 - distance_from_edge

                if y_lim[0] > y_outside:
                    arrow_start_y = arrow_proj_length + distance_from_edge
                    arrow_end_y = distance_from_edge
                else:
                    arrow_start_y = 1.0 - arrow_proj_length - distance_from_edge
                    arrow_end_y = 1.0 - distance_from_edge

                # Use figure's relative coordinates along the X and Y axis.
                tform = blended_transform_factory(ax.transAxes, ax.transAxes)

                ax.annotate(
                    "",
                    xytext=(arrow_start_x, arrow_start_y),
                    textcoords=tform,
                    xy=(arrow_end_x, arrow_end_y),
                    xycoords=tform,
                    arrowprops=dict(color=color, arrowstyle=arrow_style),
                )

        return

    def plot_line(
        self,
        ax: Axes,
        x: unyt_array,
        y: unyt_array,
        label: Union[str, None] = None,
        x_lim: Union[List, None] = None,
        y_lim: Union[List, None] = None,
        min_num_points_highlight: int = 0,
    ):
        """
        Plot a line using these parameters on some axes, x against y.

        Parameters
        ----------

        ax: Axes
            Matplotlib axes to plot on.

        x: unyt_array
            Horizontal axis data

        y: unyt_array
            Vertical axis data

        label: str
            Label associated with this data that will be included in the
            legend.

        x_lim: Union[List, None]
            A 2-length list containing the lower and upper limits of the X-axis range.

        y_lim: Union[List, None]
            A 2-length list containing the lower and upper limits of the Y-axis range.

        min_num_points_highlight: int, optional
            Minimum number of data points with the highest values of x to highlight in
            the median-line or mean-line plots.

        Notes
        -----

        If self.scatter is set to "none", this is plotted assuming the scatter
        is zero.
        """

        if not self.plot:
            return

        centers, heights, errors, additional_x, additional_y = self.create_line(
            x=x, y=y, minimum_additional_points=min_num_points_highlight
        )

        if self.scatter == "none" or errors is None:
            (line,) = ax.plot(centers, heights, label=label)
        elif self.scatter == "errorbar":
            (line, *_) = ax.errorbar(centers, heights, yerr=errors, label=label)
        elif self.scatter == "errorbar_both":
            (line, *_) = ax.errorbar(
                centers,
                heights,
                yerr=errors,
                xerr=abs(self.bins - centers),
                label=label,
                fmt=".",  # Do not plot as a line.
            )
        elif self.scatter == "shaded":
            (line,) = ax.plot(centers, heights, label=label)

            # Deal with different + and -ve errors
            if errors.shape[0]:
                if errors.ndim > 1:
                    down, up = errors
                else:
                    up = errors
                    down = errors
            else:
                up = unyt_quantity(0, units=heights.units)
                down = unyt_quantity(0, units=heights.units)

            ax.fill_between(
                centers,
                heights - down,
                heights + up,
                color=line.get_color(),
                alpha=0.3,
                linewidth=0.0,
            )

        try:
            ax.scatter(additional_x.value, additional_y.value, color=line.get_color())

            # Enter only if the plot has a valid X-axis and Y-axis ranges and there are
            # any additional data points.
            if x_lim is not None and y_lim is not None and len(additional_x) > 0:

                # Add arrows to the plot for each data point beyond X- or/and Y- axis
                # range
                self.highlight_data_outside_domain(
                    ax,
                    additional_x.value,
                    additional_y.value,
                    line.get_color(),
                    (x_lim[0].value, x_lim[1].value),
                    (y_lim[0].value, y_lim[1].value),
                )

        # In case the line object is undefined
        except NameError:
            ax.scatter(additional_x.value, additional_y.value)

        return



"""
Objects required to register derived quantities. Enables
the extension of the `catalogue` object.
"""

import numpy as np
import unyt
from typing import Union, List


class DerivedQuantities(object):
    """
    Derived quantities class. Contains methods to open (a) python
    source file(s) that create(s) derived quantities as follows:

    The source file will have access to:

    + numpy (imported as np)
    + unyt (imported as unyt)

    You should write your derived quantities as follows:

    .. code-block:: python

        self.derived_quantitiy_name = catalogue.type.field_name * 0.5
        self.derived_quantitiy_name_two = (
            catalogue.type.field_name / catalogue.type.field_name_two
        )


    For example, to register the specific star formation rate in apertures:

    .. code-block:: python

        for aperture_size in [5, 10, 30, 50, 100]:
            stellar_mass = getattr(
                catalogue.apertures,
                f"mass_star_{aperture_size}_kpc"
            )

            star_formation_rate = getattr(
                catalogue.apertures,
                f"sfr_gas_{aperture_size}_kpc"
            )

            ssfr = star_formation_rate / stellar_mass
            # Don't forget to set the name or your plot will be un-labeled!
            ssfr.name = f"Specific SFR ({aperture_size} kpc)"

            setattr(
                self, f"specific_sfr_gas_{aperture_size}_kpc", ssfr
            )


    The path to this file(s) should be passed to the __init__ method
    of this class. Note that you should only register quantities that
    you do in fact intend to use, as these are not lazily loaded
    in the same way as the properties that are built into catalogues.
    For example, in the above registration both the star formation rate
    and stellar mass are loaded from the VELOCIraptor catalogue file,
    a behaviour which may not be ideal if the catalogue is very large.

    """

    def __init__(self, registration_file_path: Union[List[str], str], catalogue):
        """
        Registers additional (derived) quantities from the
        VelociraptorCatalogue to itself, using a python
        source file that does this registration inside
        the private _register_quantities() method.

        Parameters
        ----------

        registration_file_path: Union[List[str], str]
            Path to the python source file(s). For more information
            on the contents of this file/these files, check out the information
            of this object.

        catalogue: VelociraptorCatalogue
            The catalogue to derive the quantities from.


        Returns
        -------

        DerivedQuantities
            An instance of the DerivedQuantities class with
            the properties defined in registration_file_path
            available as attributes.

        """

        if isinstance(registration_file_path, list):
            self.registration_file_paths = list(registration_file_path)
        else:
            self.registration_file_paths = [registration_file_path]
        self.catalogue = catalogue

        self._register_quantities()

        return

    def _register_quantities(self):
        """
        Registers any required derived quantities using the python
        source file defined in the initialiser.

        """

        catalogue = self.catalogue
        for file_path in self.registration_file_paths:
            with open(file_path, "r") as handle:
                exec(handle.read())

        return



"""
Tools for generating labels from catalogue datasets.
"""

import unyt
import re


def get_full_label(dataset: unyt.unyt_array):
    """
    Get the full label for one of our VelociraptorCatalogue datasets.
    This will get the automatically generated name and concatenate
    it with the _current_ untis for that dataset.
    """

    unit_tex = dataset.units.latex_representation()

    full_label = f"{dataset.name} $\\left[{unit_tex}\\right]$"

    return full_label


def get_mass_function_label_no_units(mass_function_sub_label: str):
    """
    Gets a fancy mass-function label such as:

    d$n(M_*)$/d$\log_{10}M_*$ [Mpc$^{-3}$]

    (this would be for an input of "*" and unyt.Mpc**3).
    """

    output = (
        fr"d$n(M_{mass_function_sub_label})$/d$\log_{{10}}M_{mass_function_sub_label}$"
    )

    return output


def get_mass_function_label(
    mass_function_sub_label: str, mass_function_units: unyt.Unit
):
    """
    Gets a fancy mass-function label such as:

    d$n(M_*)$/d$\log_{10}M_*$ [Mpc$^{-3}$]

    (this would be for an input of "*" and unyt.Mpc**3).
    """

    unit_repr = mass_function_units.latex_representation()
    mass_func_label = get_mass_function_label_no_units(mass_function_sub_label)

    output = fr"{mass_func_label} $\left[{unit_repr}\right]$"

    return output


def get_luminosity_function_label_no_units(luminosity_function_sub_label: str):
    """
    Gets a fancy luminosity-function label such as:

    d$n(M)$/d$M$ [Mpc$^{-3}$ mag$^{-1}$]

    (this would be for an input of "*" and unyt.Mpc**3).
    """

    output = fr"$\phi(M_{luminosity_function_sub_label})$"

    return output
